###############################################################################
##
## Zeppelin "L70" airship for FlightGear.
## Copyright (C) 2019  Franz Schmid  (Franz.Schmid(at)altmuehlnet.de)
##
##  Copyright (C) 2010 - 2016  Anders Gidenstam  (anders(at)gidenstam.org)
##  This file is licensed under the GPL license v2 or later.
##
###############################################################################

###############################################################################
# External API
#
# autoWeightOff()
# printWOW()
# shiftTrimBallast(direction, amount)
# releaseBallast(location, amount)
# refillQuickReleaseBallast(location)
# setForwardGasValves()
# setAftGasValves()
# switchEngineDirection(engine)
# about()
#

# Constants
var FORWARD = -1;
var AFT     = -2;
var FORE_BALLAST = -1;
var AFT_BALLAST = -2;
var QUICK_RELEASE_BAG_CAPACITY = 220.5; # lb
var TRIM_BAG_CAPACITY = 9000.0; # lb

var baseTrimPos = 3772;

###############################################################################
var weight_on_gear_p = "/fdm/jsbsim/forces/fbz-gear-lbs";

var trim_ballast_p =
    [
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[4]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[5]"
    ];
var fore_ballast_p =
    [
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[6]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[7]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[8]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[9]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[10]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[11]"
    ];
var fore_ballast_toggles_p =
    [
     "/controls/ballast/release[0]",
     "/controls/ballast/release[1]",
     "/controls/ballast/release[2]",
     "/controls/ballast/release[3]",
     "/controls/ballast/release[4]",
     "/controls/ballast/release[5]",
    ];
var aft_ballast_p =
    [
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[0]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[1]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[2]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[3]"
    ];
var aft_ballast_toggles_p =
    [
     "/controls/ballast/release[6]",
     "/controls/ballast/release[7]",
     "/controls/ballast/release[8]",
     "/controls/ballast/release[9]"
    ];
var fore_ballast_action_p =
    [
     "/controls/ballast/action[0]",
     "/controls/ballast/action[1]",
     "/controls/ballast/action[2]",
     "/controls/ballast/action[3]",
     "/controls/ballast/action[4]",
     "/controls/ballast/action[5]"
    ];
var aft_ballast_action_p =
    [
     "/controls/ballast/action[6]",
     "/controls/ballast/action[7]",
     "/controls/ballast/action[8]",
     "/controls/ballast/action[9]"
    ];
var all_ballast_p =
    [
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[0]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[1]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[2]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[3]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[6]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[7]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[8]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[9]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[10]",
     "/fdm/jsbsim/inertia/pointmass-weight-lbs[11]"
    ];
var all_ballast_toggles_p =
    [
     "/controls/ballast/release[0]",
     "/controls/ballast/release[1]",
     "/controls/ballast/release[2]",
     "/controls/ballast/release[3]",
     "/controls/ballast/release[4]",
     "/controls/ballast/release[5]",
     "/controls/ballast/release[6]",
     "/controls/ballast/release[7]",
     "/controls/ballast/release[8]",
     "/controls/ballast/release[9]"
    ];
var all_ballast_action_p =
    [
     "/controls/ballast/action[0]",
     "/controls/ballast/action[1]",
     "/controls/ballast/action[2]",
     "/controls/ballast/action[3]",
     "/controls/ballast/action[4]",
     "/controls/ballast/action[5]",
     "/controls/ballast/action[6]",
     "/controls/ballast/action[7]",
     "/controls/ballast/action[8]",
     "/controls/ballast/action[9]"
    ];
var current_fore_ballast_index = "/controls/ballast/current_fore_ballast_index";
var current_aft_ballast_index = "/controls/ballast/current_aft_ballast_index";

var indexTL = -1;

var printWOW = func
{
	if (getprop("/gear/gear[0]/wow"))
	{
		gui.popupTip("Current weight on gear " ~ -getprop(weight_on_gear_p) ~ " lbs.");
	}
	else
	{
		gui.popupTip("Current lift " ~ getprop("/fdm/jsbsim/static-condition/net-lift-lbs") ~ " lbs.");
	}
}

# Weight off to neutral
var autoWeighoff = func
{
	if (getprop("/gear/gear[0]/wow"))
	{
       gui.popupTip("Weight-off in progress.", 5);
	   var lift = getprop("/fdm/jsbsim/static-condition/net-lift-lbs");
	   var n = size(trim_ballast_p);
	   print("L70: Auto weigh off from " ~ (-lift) ~ " lb heavy to neutral.");
	   foreach(var p; trim_ballast_p)
	   {
		   var v = getprop(p) + lift/n;
		   interpolate(p, (v > 0 ? v : 0), 5.0);
	   }
	   settimer(func { gui.popupTip("Weight-off finished.", 5); }, 7.0);
    }
    else
       gui.popupTip("Must be on ground for weight-off.");
}

var initial_weighoff = func {
    # Set initial static condition.
    # Finding the right static condition at initialization time is tricky.
    autoWeighoff();
    settimer(autoWeighoff, 0.25);
    settimer(autoWeighoff, 1.0);
}

var init_all = func(reinit=0)
{
    if (!reinit)
    {
        # Create FG /controls/gas/ aliases for FDM owned controls.
        var fdm = "fdm/jsbsim/buoyant_forces/gas-cell";
        props.globals.getNode(gascell ~ "[0]", 1).alias(props.globals.getNode(fdm ~ "[0]/valve_open"));
        props.globals.getNode(gascell ~ "[1]", 1).alias(props.globals.getNode(fdm ~ "[1]/valve_open"));
        props.globals.getNode(gascell ~ "[2]", 1).alias(props.globals.getNode(fdm ~ "[2]/valve_open"));
        props.globals.getNode(gascell ~ "[3]", 1).alias(props.globals.getNode(fdm ~ "[3]/valve_open"));
        props.globals.getNode(gascell ~ "[4]", 1).alias(props.globals.getNode(fdm ~ "[4]/valve_open"));
        props.globals.getNode(gascell ~ "[5]", 1).alias(props.globals.getNode(fdm ~ "[5]/valve_open"));
        props.globals.getNode(gascell ~ "[6]", 1).alias(props.globals.getNode(fdm ~ "[6]/valve_open"));
        props.globals.getNode(gascell ~ "[7]", 1).alias(props.globals.getNode(fdm ~ "[7]/valve_open"));
        props.globals.getNode(gascell ~ "[8]", 1).alias(props.globals.getNode(fdm ~ "[8]/valve_open"));
        props.globals.getNode(gascell ~ "[9]", 1).alias(props.globals.getNode(fdm ~ "[9]/valve_open"));
        props.globals.getNode(gascell ~ "[10]", 1).alias(props.globals.getNode(fdm ~ "[10]/valve_open"));
        props.globals.getNode(gascell ~ "[11]", 1).alias(props.globals.getNode(fdm ~ "[11]/valve_open"));
        props.globals.getNode(gascell ~ "[12]", 1).alias(props.globals.getNode(fdm ~ "[12]/valve_open"));
        props.globals.getNode(gascell ~ "[13]", 1).alias(props.globals.getNode(fdm ~ "[13]/valve_open"));
		setprop("/controls/gas/pressurecontrol", 1);
		setprop("/controls/gas/maxfullness", 95);
    }
}

var _nordstern_initialized = 0;
setlistener("/sim/signals/fdm-initialized", func {
	var fgversion = getprop("/sim/version/flightgear");
	var (major, minor, patch) = split(".", fgversion);
	setprop("/controls/lighting/compositor", ((major >= 2020) and (minor >= 4)));
    setprop("/fdm/jsbsim/gear/unit[9]/pos-norm", 0);
	if (getprop("/sim/model/L70/turn-into-wind"))
	{
		setprop("/orientation/heading-deg", getprop("/environment/metar/base-wind-dir-deg"));
	}
    ground_crew.init();
    init_all(_nordstern_initialized);
    _nordstern_initialized = 1;
	# disable default blackout system
	setprop("/sim/rendering/redout/enabled", 0);
	for (var n = 0; n < 100; n = n + 1)
	{
		var pRule = getprop("/sim/systems/property-rule[" ~ n ~ "]/name");
		if (pRule == nil)
		{
			break;
		}
		if (pRule == "g force effects")
		{
			setprop("/sim/systems/property-rule[" ~ n ~ "]/serviceable", 0);
			break;
		}
	}
	if (getprop("/controls/lighting/compositor"))
	{
#	Get index of Flashlight
		var lightNode = props.globals.getNode("/scenery/lights").getChildren();
		forindex(i; lightNode)
		{
			if (lightNode[i].getValue("name") == "Taschenlampe")
			{
				indexTL = i;
				break;
			}
		}
		setprop("controls/lighting/flash-light-index", indexTL);
		props.globals.getNode("controls/lighting/flash-light-x").alias(props.globals.getNode("/scenery/lights/light[" ~ indexTL ~ "]/position/x-m"));
		props.globals.getNode("controls/lighting/flash-light-y").alias(props.globals.getNode("/scenery/lights/light[" ~ indexTL ~ "]/position/y-m"));
		props.globals.getNode("controls/lighting/flash-light-z").alias(props.globals.getNode("/scenery/lights/light[" ~ indexTL ~ "]/position/z-m"));
		props.globals.getNode("controls/lighting/flash-light-h").alias(props.globals.getNode("/scenery/lights/light[" ~ indexTL ~ "]/direction/heading-deg"));
		props.globals.getNode("controls/lighting/flash-light-p").alias(props.globals.getNode("/scenery/lights/light[" ~ indexTL ~ "]/direction/pitch-deg"));
		props.globals.getNode("controls/lighting/flash-light-r").alias(props.globals.getNode("/scenery/lights/light[" ~ indexTL ~ "]/direction/roll-deg"));
	}
    props.globals.initNode("/sim/rendering/als-secondary-lights/use-landing-light-ext",0,"BOOL");
    props.globals.initNode("/sim/rendering/als-secondary-lights/use-taxi-light-ext",0,"BOOL");
	# initialize tanks with proper values
	var lbs = getprop("/consumables/fuel/total-fuel-lbs") / 14.0;
	var gals = lbs / getprop("/consumables/fuel/tank[0]/density-ppg");
	for (var t = 0; t < 14; t = t + 1)
	{
		setprop("/consumables/fuel/tank[" ~ t ~ "]/capacity-gal_us", gals);
		setprop("/consumables/fuel/tank[" ~ t ~ "]/level-gal_us", gals);
		setprop("/consumables/fuel/tank[" ~ t ~ "]/level-lbs", lbs);
		setprop("/consumables/fuel/tank[" ~ t ~ "]/level-norm", 1);
	}
    setprop("controls/lighting/landing-light", 0);
	setprop("/controls/engines/helpengine-lever", 0);
	setprop("systems/electrical/mains", 0);
	setprop("/fdm/jsbsim/crew/helmsman/heading-magnetic-cmd-deg" , int(getprop("/orientation/heading-deg")));
	setprop("/fdm/jsbsim/crew/elevatorman/altitude-cmd-ft" , getprop("/position/altitude-ft"));
	setprop("/autopilot/settings/target-altitude-m", getprop("/autopilot/settings/target-altitude-ft") * 0.3048);
	setprop("/autopilot/settings/true-heading-deg" , int(getprop("/orientation/heading-deg")));
	setprop("/autopilot/settings/target-pitch-deg" , 0);
	setprop("/autopilot/settings/target-speed-kt" , 0);
	setprop("/autopilot/settings/target-speed-kmh", 0);
	setprop("/autopilot/settings/autoland", 0);
	setprop("/autopilot/settings/target-cgx" , 3772.44);
	setprop("/autopilot/internal/throttle-cmd" , 0);
	setprop("/autopilot/internal/lift_diff" , 0);
	setprop("/autopilot/internal/lift_size" , 1);
	setprop("/autopilot/settings/vertical-speed-fps" , 0);
	setprop("/autopilot/settings/target-distance-m" , 4000);
	setprop("/controls/engines/opselreq", -1);
    setprop("/controls/engines/opstatereq", 0);
    setprop("/controls/engines/opstateresp", 0);
    setprop("/controls/engines/engine[0]/opstatereq", 0);
    setprop("/controls/engines/engine[0]/opstateresp", 0);
    setprop("/controls/engines/engine[0]/opstatesrc", 0);
    setprop("/controls/engines/engine[1]/opstatereq", 0);
    setprop("/controls/engines/engine[1]/opstateresp", 0);
    setprop("/controls/engines/engine[1]/opstatesrc", 0);
    setprop("/controls/engines/engine[2]/opstatereq", 0);
    setprop("/controls/engines/engine[2]/opstateresp", 0);
    setprop("/controls/engines/engine[2]/opstatesrc", 0);
    setprop("/controls/engines/engine[3]/opstatereq", 0);
    setprop("/controls/engines/engine[3]/opstateresp", 0);
    setprop("/controls/engines/engine[3]/opstatesrc", 0);
    setprop("/controls/engines/engine[4]/opstatereq", 0);
    setprop("/controls/engines/engine[4]/opstateresp", 0);
    setprop("/controls/engines/engine[4]/opstatesrc", 0);
    setprop("/controls/engines/engine[5]/opstatereq", 0);
    setprop("/controls/engines/engine[5]/opstateresp", 0);
    setprop("/controls/engines/engine[5]/opstatesrc", 0);
	setprop("/controls/engines/engine[0]/defect", 1);
	setprop("/controls/engines/engine[1]/defect", 1);
	setprop("/controls/engines/engine[2]/defect", 1);
	setprop("/controls/engines/engine[3]/defect", 1);
	setprop("/controls/engines/engine[4]/defect", 1);
	setprop("/controls/engines/engine[5]/defect", 1);
    setprop("/controls/engines/clustatereq", 0);
    setprop("/controls/engines/clustateresp", 0);
    setprop("/controls/engines/engine[0]/clustatereq", 0);
    setprop("/controls/engines/engine[0]/clustateresp", 0);
    setprop("/controls/engines/engine[1]/clustatereq", 0);
    setprop("/controls/engines/engine[1]/clustateresp", 0);
    setprop("/controls/engines/engine[2]/clustatereq", 0);
    setprop("/controls/engines/engine[2]/clustateresp", 0);
    setprop("/controls/engines/engine[3]/clustatereq", 0);
    setprop("/controls/engines/engine[3]/clustateresp", 0);
    setprop("/controls/engines/engine[4]/clustatereq", 0);
    setprop("/controls/engines/engine[4]/clustateresp", 0);
    setprop("/controls/engines/engine[5]/clustatereq", 0);
    setprop("/controls/engines/engine[5]/clustateresp", 0);
    setprop("/controls/ballast/action[0]", 0);
    setprop("/controls/ballast/action[1]", 0);
    setprop("/controls/ballast/action[2]", 0);
    setprop("/controls/ballast/action[3]", 0);
    setprop("/controls/ballast/action[4]", 0);
    setprop("/controls/ballast/action[5]", 0);
    setprop("/controls/ballast/action[6]", 0);
    setprop("/controls/ballast/action[7]", 0);
    setprop("/controls/ballast/action[8]", 0);
    setprop("/controls/ballast/action[9]", 0);
    setprop("/controls/ballast/current_fore_ballast_index", 0);
    setprop("/controls/ballast/current_aft_ballast_index", 0);
    setprop("controls/gas/overpressure[0]", 0);
    setprop("controls/gas/overpressure[1]", 0);
    setprop("controls/gas/overpressure[2]", 0);
    setprop("controls/gas/overpressure[3]", 0);
    setprop("controls/gas/overpressure[4]", 0);
    setprop("controls/gas/overpressure[5]", 0);
    setprop("controls/gas/overpressure[6]", 0);
    setprop("controls/gas/overpressure[7]", 0);
    setprop("controls/gas/overpressure[8]", 0);
    setprop("controls/gas/overpressure[9]", 0);
    setprop("controls/gas/overpressure[10]", 0);
    setprop("controls/gas/overpressure[11]", 0);
    setprop("controls/gas/overpressure[12]", 0);
    setprop("controls/gas/overpressure[13]", 0);
    setprop("controls/gas/destroypressure[0]", 0);
    setprop("controls/gas/destroypressure[1]", 0);
    setprop("controls/gas/destroypressure[2]", 0);
    setprop("controls/gas/destroypressure[3]", 0);
    setprop("controls/gas/destroypressure[4]", 0);
    setprop("controls/gas/destroypressure[5]", 0);
    setprop("controls/gas/destroypressure[6]", 0);
    setprop("controls/gas/destroypressure[7]", 0);
    setprop("controls/gas/destroypressure[8]", 0);
    setprop("controls/gas/destroypressure[9]", 0);
    setprop("controls/gas/destroypressure[10]", 0);
    setprop("controls/gas/destroypressure[11]", 0);
    setprop("controls/gas/destroypressure[12]", 0);
    setprop("controls/gas/destroypressure[13]", 0);
	setprop("/controls/flight/trim", 0);
	setprop("/systems/electrical/outputs/adf", 24.0);
	setprop("/systems/electrical/outputs/comm[0]", 24.0);
    setlistener("controls/ballast/action[0]", func() { broadcast.send(message_id["drop_water"] ~ Binary.encodeByte(0) ~ Binary.encodeInt(getprop("/controls/ballast/action[0]"))); });
    setlistener("controls/ballast/action[1]", func() { broadcast.send(message_id["drop_water"] ~ Binary.encodeByte(1) ~ Binary.encodeInt(getprop("/controls/ballast/action[1]"))); });
    setlistener("controls/ballast/action[2]", func() { broadcast.send(message_id["drop_water"] ~ Binary.encodeByte(2) ~ Binary.encodeInt(getprop("/controls/ballast/action[2]"))); });
    setlistener("controls/ballast/action[3]", func() { broadcast.send(message_id["drop_water"] ~ Binary.encodeByte(3) ~ Binary.encodeInt(getprop("/controls/ballast/action[3]"))); });
    setlistener("controls/ballast/action[4]", func() { broadcast.send(message_id["drop_water"] ~ Binary.encodeByte(4) ~ Binary.encodeInt(getprop("/controls/ballast/action[4]"))); });
    setlistener("controls/ballast/action[5]", func() { broadcast.send(message_id["drop_water"] ~ Binary.encodeByte(5) ~ Binary.encodeInt(getprop("/controls/ballast/action[5]"))); });
    setlistener("controls/ballast/action[6]", func() { broadcast.send(message_id["drop_water"] ~ Binary.encodeByte(6) ~ Binary.encodeInt(getprop("/controls/ballast/action[6]"))); });
    setlistener("controls/ballast/action[7]", func() { broadcast.send(message_id["drop_water"] ~ Binary.encodeByte(7) ~ Binary.encodeInt(getprop("/controls/ballast/action[7]"))); });
    setlistener("controls/ballast/action[8]", func() { broadcast.send(message_id["drop_water"] ~ Binary.encodeByte(8) ~ Binary.encodeInt(getprop("/controls/ballast/action[8]"))); });
    setlistener("controls/ballast/action[9]", func() { broadcast.send(message_id["drop_water"] ~ Binary.encodeByte(9) ~ Binary.encodeInt(getprop("/controls/ballast/action[9]"))); });
    setlistener("/sim/weapons/dropping", func() {
               broadcast.send(message_id["drop_bombs"] ~ Binary.encodeByte(getprop("/sim/weapons/bomb[0]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[1]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[2]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[3]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[4]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[5]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[6]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[7]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[8]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[9]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[10]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[11]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[12]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[13]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[14]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[15]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[16]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[17]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[18]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[19]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[20]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[21]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[22]/present"))
                                                       ~ Binary.encodeByte(getprop("/sim/weapons/bomb[23]/present"))
                             ); });
   settimer(func {
                ground_crew.place_ground_crew(geo.aircraft_position(), getprop("/orientation/heading-deg"));
                ground_crew.activate();
            }, 1.0);
    print("L70 Systems ... Check");

    setlistener("/sim/signals/click", func {
        var click_pos = geo.click_position();
        if (__kbd.ctrl.getBoolValue()) {
            L70.ground_crew.place_ground_crew(click_pos, nil, click_pos.alt());
			setprop("/sim/messages/ground", "Ground Crew placed.");
        }
    });

	# destroy propeller 1 and 6 if agl too low
    setlistener("/gear/gear[7]/wow", func {
		if (getprop("/gear/gear[7]/wow"))
		{
			if (getprop("/engines/engine[0]/running") and (getprop("/fdm/jsbsim/propulsion/engine[0]/blade-angle") != 0))
			{
				setprop("/controls/engines/engine[0]/defect", 0);
				SetEngineState(0, 0);
				setprop("/sim/messages/ground", "Fore Propeller destroyed!");
			}
		}
    });
    setlistener("/gear/gear[8]/wow", func {
		if (getprop("/gear/gear[8]/wow"))
		{
			if (getprop("/engines/engine[5]/running") and (getprop("/fdm/jsbsim/propulsion/engine[5]/blade-angle") != 0))
			{
				setprop("/controls/engines/engine[5]/defect", 0);
				SetEngineState(5, 0);
				setprop("/sim/messages/ground", "Aft Propeller destroyed!");
			}
		}
    });
});

# Initialize scenario network for full participation.
#io.load_nasal(getprop("/sim/aircraft-dir") ~ "/Nasal/scenario-network.nas", "L70");
L70.scenario_network_init(1);

# Release one or more quick-release ballast units.
var releaseBallast = func(location, amount)
{
   if (getprop("/fdm/jsbsim/systems/doors/BaDoor/position-norm") < 1)
   {
      gui.popupTip("open doors first.");
      return;
   }
    var units   = nil;
    var toggles = nil;
    var action = nil;
    var sIndex = nil;
    if (location == FORE_BALLAST) {
        units   = fore_ballast_p;
        toggles = fore_ballast_toggles_p;
        action = fore_ballast_action_p;
        sIndex = getprop(current_fore_ballast_index);
    } elsif (location == AFT_BALLAST) {
        units   = aft_ballast_p;
        toggles = aft_ballast_toggles_p;
        action = aft_ballast_action_p;
        sIndex = getprop(current_aft_ballast_index);
    } else {
        printlog("warn",
                 "L70.releaseBallast(" ~ location ~ ", " ~ amount ~
                 "): Invalid ballast location.");
        return;
    }
    forindex (var i; units)
    {
        if (!amount) return;
        if ((getprop(action[sIndex]) == 0) and (getprop(units[sIndex]) > 0.0))
        {
            var act = action[sIndex];
            setprop(act, 1);
            setprop(toggles[sIndex], 1.0);
            amount -= 1;
            sIndex += 1;
        }
        if (location == FORE_BALLAST)
        {
           if (sIndex > 5)
           {
               sIndex = 0;
               gui.popupTip("Fore Ballast empty!");
           }
           if ((sIndex == 0) and (getprop(units[sIndex]) == 0))
           {
               gui.popupTip("Fore Ballast empty!");
           }
           setprop(current_fore_ballast_index, sIndex);
        }
        else
        {
           if (sIndex > 3)
           {
               sIndex = 0;
               gui.popupTip("Aft Ballast empty!");
           }
           if ((sIndex == 0) and (getprop(units[sIndex]) == 0))
           {
               gui.popupTip("Aft Ballast empty!");
           }
           setprop(current_aft_ballast_index, sIndex);
        }
    }
}

var stopBallastAction = func(prop)
{
   setprop(prop, 0);
}

# Refills empty "ballasthosen" from the trim ballast bags.
var refillQuickReleaseBallast = func()
{
    var units   = all_ballast_p;
    var toggles = all_ballast_toggles_p;
	var action  = all_ballast_action_p;
    setprop(current_fore_ballast_index, 0);
    setprop(current_aft_ballast_index, 0);
    var n = size(trim_ballast_p);
    forindex (var ab; action)
    {
		setprop(action[ab], 0);
	}
	var val = 0;
	var count = 0;
    forindex (var qb; units)
    {
		if (getprop(units[qb]) != QUICK_RELEASE_BAG_CAPACITY)
		{
			count = count + 1;
			val = val + (QUICK_RELEASE_BAG_CAPACITY - getprop(units[qb]));
		}
		setprop(toggles[qb], 0.0);
	}
	if (count == 0)
	{
		return;
	}
	val = val / n;
	var last = 0;
    foreach(var tb; trim_ballast_p)
    {
		var tmp = getprop(tb) - val;
		if (tmp < 0)
		{
			last = last + tmp;
			tmp = 0;
			gui.popupTip("Trim Ballast empty");
		}
	    interpolate(tb, tmp, 10.0);
	}
	var subtract = last / count;
    forindex (var db; units)
    {
		if (getprop(units[db]) < QUICK_RELEASE_BAG_CAPACITY)
		{
			interpolate(units[db], QUICK_RELEASE_BAG_CAPACITY + subtract, 10.0);
		}
	}
    settimer(func { gui.popupTip("Refill Ballast finished\n" ~ getprop("/fdm/jsbsim/static-condition/net-weight") * 0.453 ~ " kg left") }, 10);
}

# Listen to the /controls/ballast/release[x] controls.
forindex(var i; fore_ballast_toggles_p)
{
    setlistener(fore_ballast_toggles_p[i], func (n) {
        if ((getprop(fore_ballast_toggles_p[n.getIndex()]) == 1) and (getprop(fore_ballast_p[n.getIndex()]) > 0))
        {
           setprop(fore_ballast_action_p[n.getIndex()], 1);
           settimer(func { stopBallastAction(fore_ballast_action_p[n.getIndex()]); }, 3);
        }
        interpolate(fore_ballast_p[n.getIndex()], 0.0, 1.0);
    }, 0, 0);
}
forindex(var i; aft_ballast_toggles_p)
{
    setlistener(aft_ballast_toggles_p[i], func (n) {
        if ((getprop(aft_ballast_toggles_p[n.getIndex() - size(fore_ballast_toggles_p)]) == 1) and (getprop(aft_ballast_p[n.getIndex() - size(fore_ballast_toggles_p)]) > 0))
        {
		   setprop(aft_ballast_action_p[n.getIndex() - size(fore_ballast_toggles_p)], 1);
           settimer(func { stopBallastAction(aft_ballast_action_p[n.getIndex() - size(fore_ballast_toggles_p)]); }, 3);
        }
        interpolate(aft_ballast_p[n.getIndex() - size(fore_ballast_toggles_p)], 0.0, 1.0);
    }, 0, 0);
}


###############################################################################
# Gas valve controls
var gascell = "controls/gas/valve-cmd-norm";

var setForwardGasValves = func (v) {
    setprop(gascell ~ "[13]", v);
    setprop(gascell ~ "[12]", v);
    setprop(gascell ~ "[11]", v);
    setprop(gascell ~ "[10]", v);
    setprop(gascell ~ "[9]", v);
    setprop(gascell ~ "[8]", v);
    setprop(gascell ~ "[7]", v);
}

var setAftGasValves = func (v) {
    setprop(gascell ~ "[6]", v);
    setprop(gascell ~ "[5]", v);
    setprop(gascell ~ "[4]", v);
    setprop(gascell ~ "[3]", v);
    setprop(gascell ~ "[2]", v);
    setprop(gascell ~ "[1]", v);
    setprop(gascell ~ "[0]", v);
}

var elevator_trim_fwd = func()
{
	var trim = getprop("/controls/flight/trim");
	if (trim < -1.0)
	{
		trim = -1.0;
		setprop("/controls/flight/trim", trim);
	}
	if (trim > 1.0)
	{
		trim = 1.0;
		setprop("/controls/flight/trim", trim);
	}
	var lift = baseTrimPos - (trim * 4000);
    setprop("/fdm/jsbsim/inertia/pointmass-location-X-inches[16]", lift);
}

###############################################################################
# Engine controls.

var GetTankState = func()
{
	var result = 0;
	for (var i = 0; i < 14; i = i + 1)
	{
		if (!getprop("/consumables/fuel/tank[" ~ i ~ "]/empty"))
		{
			result = 1;
			break;
		}
	}
	return result;
}

var SetEngineState = func (eng, state)
{
	var engineJSB = "/fdm/jsbsim/propulsion/engine" ~ "[" ~ eng ~ "]";
	var engineFG  = "/engines/engine" ~ "[" ~ eng ~ "]";
	var engineCon = "/controls/engines/engine" ~ "[" ~ eng ~ "]";
	var dir       = engineJSB ~ "/yaw-angle-rad";
	if ((GetTankState() == 0) or (state == 0) or (getprop("/controls/engines/engine[" ~ eng ~ "]/defect") == 0))
	{
		setprop(engineCon ~ "/cutoff", 1);
		setprop(engineCon ~ "/throttle", 0.0);
		setprop(engineCon ~ "/magnetos", 0);
		setprop(engineCon ~ "/starter", 0);
		setprop(dir, 0);
		setprop(engineCon ~ "/opstatereq", 0);
		setprop("/controls/engines/opstatereq", state);
		settimer(func {
				setprop(engineCon ~ "/opstateresp", 0);
				setprop("/controls/engines/opstateresp", 0);
				setprop("/controls/engines/opstatereq", 0);
				if (eng == 0)
				{
					if (getprop("controls/engines/helpengine-lever") > 0.5)
					{
						interpolate("systems/electrical/mains", 1, 2.0);
					}
					else
					{
						interpolate("systems/electrical/mains", 0, 2.0);
					}
				}
			}, 3);
		return;
	}
	if (getprop(engineFG ~ "/running"))
	{
		setprop(engineCon ~ "/starter", 0);
		setprop(dir, 0);
		setprop(engineCon ~ "/opstatereq", state);
		setprop("/controls/engines/opstatereq", state);
		if (state == -1)
		{
			setprop(engineCon ~ "/throttle", 0.6);
			setprop(dir, 3.14159265);
		}
		if (state == -2)
		{
			setprop(engineCon ~ "/throttle", 0.9);
			setprop(dir, 3.14159265);
		}
		if (state == 1)
		{
			setprop(engineCon ~ "/throttle", 0.05);
		}
		if (state == 2)
		{
			setprop(engineCon ~ "/throttle", 0.6);
		}
		if (state == 3)
		{
			setprop(engineCon ~ "/throttle", 0.85);
		}
		if (state == 4)
		{
			setprop(engineCon ~ "/throttle", 1.0);
		}
		settimer(func { setprop(engineCon ~ "/opstateresp", state);
			setprop("/controls/engines/opstateresp", state);
			setprop("/controls/engines/opstatereq", state); }, 3);
	}
	else
	{
		if (getprop("systems/electrical/mains") < 1)
		{
			gui.popupTip("Start help engine first.");
			setprop(engineCon ~ "/opstatereq", 0);
			setprop(engineCon ~ "/opstateresp", 0);
			setprop("/controls/engines/opstatereq", 0);
			return;
		}
		if (state != 0)
		{
			setprop(engineCon ~ "/cutoff", 0);
			setprop(engineCon ~ "/throttle", 0.9);
			setprop(engineCon ~ "/magnetos", 3);
			setprop(engineCon ~ "/opstatereq", 1);
			setprop(engineCon ~ "/opstateresp", 0);
			setprop("/controls/engines/opstatereq", state);
			setprop(engineCon ~ "/mixture", getprop("/fdm/jsbsim/systems/engine/mixture-min"));
			setprop(dir, 0);
			setprop(engineCon ~ "/starter", 1);
			var engineRunning = setlistener(engineFG ~ "/running", func {
				if (getprop(engineFG ~ "/running"))
				{
					removelistener(engineRunning);
					if (eng == 0)
					{
						interpolate("systems/electrical/mains", 2, 2.0);
					}
					setprop(engineCon ~ "/starter", 0);
					setprop(engineCon ~ "/opstatereq", state);
					setprop(engineCon ~ "/opstateresp", state);
					setprop("/controls/engines/opstateresp", state);
					setprop("/controls/engines/opstatereq", state);
					if (state == -1)
					{
						setprop(engineCon ~ "/throttle", 0.6);
						setprop(dir, 3.14159265);
					}
					if (state == -2)
					{
						setprop(engineCon ~ "/throttle", 0.9);
						setprop(dir, 3.14159265);
					}
					if (state == 1)
					{
						setprop(engineCon ~ "/throttle", 0.05);
					}
					if (state == 2)
					{
						setprop(engineCon ~ "/throttle", 0.6);
					}
					if (state == 3)
					{
						setprop(engineCon ~ "/throttle", 0.85);
					}
					if (state == 4)
					{
						setprop(engineCon ~ "/throttle", 1.0);
					}
				}
			});
		}
	}
}

var SetEngineStateGondola = func (eng, state)
{
	setprop("/controls/engines/engine[" ~ eng ~ "]/opstatesrc", 1);
	SetEngineState(eng, state);
	setprop("/controls/engines/engine[" ~ eng ~ "]/opstatesrc", 0);
}

var SetEngineStateTele = func (eng, state)
{
	setprop("/controls/engines/opstatereq", state);
	if (eng != -1)
	{
		setprop("/controls/engines/engine[" ~ eng ~ "]/opstatesrc", 0);
		SetEngineState(eng, state);
	}
	else
	{
		for (var i = 0; i < 6; i = i + 1)
		{
			setprop("/controls/engines/engine[" ~ i ~ "]/opstatesrc", 0);
			SetEngineState(i, state);
		}
	}
}

var selectEngine = func(eng)
{
	setprop("/controls/engines/opselreq", eng);
	settimer(func {
					setprop("/fdm/jsbsim/propulsion/active_engine", eng);
					if (eng != -1)
					{
						setprop("/controls/engines/opstateresp", getprop("/controls/engines/engine[" ~ eng ~ "]/opstateresp"));
						setprop("/controls/engines/opstatereq", getprop("/controls/engines/engine[" ~ eng ~ "]/opstatereq"));
						setprop("/controls/engines/clustateresp", getprop("/controls/engines/engine[" ~ eng ~ "]/clustateresp"));
						setprop("/controls/engines/clustatereq", getprop("/controls/engines/engine[" ~ eng ~ "]/clustatereq"));
					}
				  }, 3);
}

var toggleClutch = func(eng)
{
	var target = 0;
	if (getprop("/controls/engines/engine" ~ "[" ~ eng ~ "]/clutch") > 0)
	{
		setprop("/controls/engines/engine" ~ "[" ~ eng ~ "]/clustatereq", 0);
		target = 0;
		settimer(func { setprop("/controls/engines/engine" ~ "[" ~ eng ~ "]/clustateresp", 0);
					  }, 3);
	}
	else
	{
		setprop("/controls/engines/engine" ~ "[" ~ eng ~ "]/clustatereq", 1);
		target = 1;
		settimer(func { setprop("/controls/engines/engine" ~ "[" ~ eng ~ "]/clustateresp", 1);
					  }, 3);
	}
	interpolate("/controls/engines/engine" ~ "[" ~ eng ~ "]/clutch", target, 2.0);
}

var engageClutch = func()
{
	gui.popupTip("Clutch engaged");
	var eng = getprop("/fdm/jsbsim/propulsion/active_engine");
	if (eng != -1)
	{
		setprop("/controls/engines/clustatereq", 1);
		setprop("/controls/engines/engine" ~ "[" ~ eng ~ "]/clustatereq", 1);
		settimer(func { setprop("/controls/engines/engine" ~ "[" ~ eng ~ "]/clustateresp", 1);
						setprop("/controls/engines/clustateresp", 1);
					  }, 3);
		interpolate("/controls/engines/engine" ~ "[" ~ eng ~ "]/clutch", 1, 2.0);
	}
	else
	{
		for (var i = 0; i < 6; i = i + 1)
		{
			setprop("/controls/engines/clustatereq", 1);
			setprop("/controls/engines/engine" ~ "[" ~ i ~ "]/clustatereq", 1);
			interpolate("/controls/engines/engine" ~ "[" ~ i ~ "]/clutch", 1, 2.0);
		}
		settimer(func { for (var i = 0; i < 6; i = i + 1)
						{
							setprop("/controls/engines/engine" ~ "[" ~ i ~ "]/clustateresp", 1);
						}
						setprop("/controls/engines/clustateresp", 1);
					  }, 3);
	}
}

var disEngageClutch = func()
{
	gui.popupTip("Clutch disengaged");
	var eng = getprop("/fdm/jsbsim/propulsion/active_engine");
	if (eng != -1)
	{
		setprop("/controls/engines/clustatereq", 0);
		setprop("/controls/engines/engine" ~ "[" ~ eng ~ "]/clustatereq", 0);
		settimer(func { setprop("/controls/engines/engine" ~ "[" ~ eng ~ "]/clustateresp", 0);
						setprop("/controls/engines/clustateresp", 0);
					  }, 3);
		interpolate("/controls/engines/engine" ~ "[" ~ eng ~ "]/clutch", 0, 1.0);
	}
	else
	{
		for (var i = 0; i < 6; i = i + 1)
		{
			setprop("/controls/engines/clustatereq", 0);
			setprop("/controls/engines/engine" ~ "[" ~ i ~ "]/clustatereq", 0);
			interpolate("/controls/engines/engine" ~ "[" ~ i ~ "]/clutch", 0, 1.0);
		}
		settimer(func { for (var i = 0; i < 6; i = i + 1)
						{
							setprop("/controls/engines/engine" ~ "[" ~ i ~ "]/clustateresp", 0);
						}
						setprop("/controls/engines/clustateresp", 0);
					  }, 3);
	}
}

var incThrottle = func()
{
	var eng = getprop("/fdm/jsbsim/propulsion/active_engine");
	if (eng != -1)
	{
		setprop("/controls/engines/engine[" ~ eng ~ "]/opstatesrc", 1);
		var engineCon = "/controls/engines/engine" ~ "[" ~ eng ~ "]";
		if ((getprop(engineCon ~ "/opstatereq") == getprop(engineCon ~ "/opstateresp")) and (getprop(engineCon ~ "/opstateresp") < 4))
		{
			var state = getprop(engineCon ~ "/opstateresp") + 1;
			SetEngineState(eng, state);
		}
	}
	else
	{
		if (getprop("systems/electrical/mains") < 1.5)
		{
			setprop("controls/engines/helpengine-lever", 1);
			setprop("systems/electrical/mains", 2);
		}
		for (var i = 0; i < 6; i = i + 1)
		{
			var engineCon = "/controls/engines/engine" ~ "[" ~ i ~ "]";
			setprop(engineCon ~ "/opstatesrc", 0);
			if ((getprop(engineCon ~ "/opstatereq") == getprop(engineCon ~ "/opstateresp")) and (getprop(engineCon ~ "/opstateresp") < 4))
			{
				var state = getprop(engineCon ~ "/opstateresp") + 1;
				SetEngineState(i, state);
			}
		}
	}
}

var decThrottle = func()
{
	var eng = getprop("/fdm/jsbsim/propulsion/active_engine");
	if (eng != -1)
	{
		setprop("/controls/engines/engine[" ~ eng ~ "]/opstatesrc", 1);
		var engineCon = "/controls/engines/engine" ~ "[" ~ eng ~ "]";
		if ((getprop(engineCon ~ "/opstatereq") == getprop(engineCon ~ "/opstateresp")) and (getprop(engineCon ~ "/opstateresp") > -2))
		{
			var state = getprop(engineCon ~ "/opstateresp") - 1;
			SetEngineState(eng, state);
		}
	}
	else
	{
		if (getprop("/controls/engines/engine[0]/opstateresp") - 1 == 0)
		{
			if (getprop("controls/engines/helpengine-lever") < 0.5)
			{
				setprop("systems/electrical/mains", 0);
			}
			else
			{
				setprop("systems/electrical/mains", 1);
			}
		}
		for (var i = 0; i < 6; i = i + 1)
		{
			var engineCon = "/controls/engines/engine" ~ "[" ~ i ~ "]";
			setprop(engineCon ~ "/opstatesrc", 0);
			if ((getprop(engineCon ~ "/opstatereq") == getprop(engineCon ~ "/opstateresp")) and (getprop(engineCon ~ "/opstateresp") > -2))
			{
				var state = getprop(engineCon ~ "/opstateresp") - 1;
				SetEngineState(i, state);
			}
		}
	}
}

var callDeferredFunc = func(f, p)
{
	if (getprop("/sim/rendering/redout/alpha") > 0.2)
	{
		settimer(func { call(f, p); }, getprop("/sim/rendering/redout/alpha"));
	}
	else
	{
		call(f, p);
	}
}

var EnableHelmsMan = func (state)
{
   setprop("/fdm/jsbsim/crew/helmsman/enabled" , state);
}

var selectNextBomb = func()
{
	if (getprop("systems/electrical/mains") > 0)
	{
		var found = 0;
		for (var i = 0; i < 24; i = i + 1)
		{
			if ((getprop("/sim/weapons/bomb[" ~ i ~ "]/tobedropped") == 0) and (getprop("/sim/weapons/bomb[" ~ i ~ "]/dropped") == 0) and (getprop("/sim/weapons/bomb[" ~ i ~ "]/present") == 1))
			{
				prepareBombs(i);
				found = 1;
				gui.popupTip("bomb selected");
				break;
			}
		}
		if (found == 0)
		{
			gui.popupTip("no more bombs available!");
		}
	}
}

var prepareBombs = func(num)
{
   if (getprop("/sim/weapons/bomb[" ~ num ~ "]/tobedropped") == 0)
   {
      setprop("/sim/weapons/bomb[" ~ num ~ "]/tobedropped", 1);
   }
   else
   {
      setprop("/sim/weapons/bomb[" ~ num ~ "]/tobedropped", 0);
   }
}

var dropBombs = func()
{
	if (getprop("systems/electrical/mains") > 0)
	{
		setprop("/sim/weapons/dropping", 1);
		settimer(dropBombsWork, 0.25);
	}
}

var dropBombsWork = func()
{
   if (getprop("fdm/jsbsim/systems/doors/BDoor/position-norm") < 1)
   {
      gui.popupTip("open doors first.");
      setprop("/sim/weapons/dropping", 0);
      return;
   }
   if (getprop("/position/altitude-agl-ft") < 400)
   {
      gui.popupTip("altitude too low.");
      setprop("/sim/weapons/dropping", 0);
      return;
   }
   var dropDone = 0;
   for (var i = 0; i < 24; i = i + 1)
   {
      if ((getprop("/sim/weapons/bomb[" ~ i ~ "]/tobedropped") == 1) and (getprop("/sim/weapons/bomb[" ~ i ~ "]/dropped") == 0) and (getprop("/sim/weapons/bomb[" ~ i ~ "]/present") == 1))
      {
         dropDone = 1;
         setprop("/sim/weapons/bomb[" ~ i ~ "]/dropped", 1);
         setprop("/sim/weapons/bomb[" ~ i ~ "]/present", 0);
         if (i < 8)
         {
            setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[23]", getprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[23]") - 220);
         }
         else 
         {
            if (i > 15)
            {
               setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[25]", getprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[25]") - 220);
            }
            else
            {
               setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[24]", getprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[24]") - 220);
            }
         }
      }
   }
   if (dropDone == 1)
   {
      gui.popupTip("bombs dropped.");
   }
   setprop("/sim/weapons/dropping", 0);
}

var impactBomb = func(n)
{
   var node = props.globals.getNode(n.getValue(), 1);
   var impact = n.getValue();
   var solid = getprop(impact ~ "/material/solid");
   if (solid)
   {
      var lat = node.getNode("impact/latitude-deg").getValue();
      var lon = node.getNode("impact/longitude-deg").getValue();
      var ele = node.getNode("impact/elevation-m").getValue();
      var hea = node.getNode("impact/heading-deg").getValue();
      geo.put_model("Aircraft/L70/Models/crater.xml", lat, lon, ele + 0.1, hea, 0, 0);
      geo.put_model("Aircraft/L70/Models/bombe.xml", lat, lon, ele + 0.25, hea, 0, 0);
      var pos = geo.Coord.new().set_latlon(lat, lon, ele);
      start_terrain_fire(lat, lon, ele + 0.25, 230);
      broadcast.send(message_id["bomb_impact"] ~ Binary.encodeCoord(pos));
   }
}

setlistener("sim/ai/aircraft/impact/bomb", impactBomb);

var toggleIntoWind = func()
{
	var aLock = getprop("/autopilot/locks/speed");
	if (aLock == "wind")
	{
		setprop("/fdm/jsbsim/crew/helmsman/enabled" , 1);
		setprop("/autopilot/locks/speed", "");
		SetEngineState(1, 1);
		SetEngineState(2, 1);
		SetEngineState(3, 1);
		SetEngineState(4, 1);
	}
	else if (aLock == "speed-with-throttle")
	{
		SetEngineState(1, 1);
		SetEngineState(2, 1);
		SetEngineState(3, 1);
		SetEngineState(4, 1);
		setprop("/fdm/jsbsim/crew/helmsman/enabled" , 1);
		setprop("/autopilot/locks/heading-enable", 1);
		setprop("/autopilot/locks/heading-dg-hold", 1);
		setprop("/autopilot/locks/heading-true-hold", 0);
		setprop("/autopilot/locks/speed", "wind");
	}
	else
	{
		SetEngineState(0, 0);
		SetEngineState(1, 1);
		SetEngineState(2, 1);
		SetEngineState(3, 1);
		SetEngineState(4, 1);
		SetEngineState(5, 0);
		setprop("/fdm/jsbsim/crew/helmsman/enabled" , 1);
		setprop("/autopilot/locks/heading-enable", 1);
		setprop("/autopilot/locks/heading-dg-hold", 1);
		setprop("/autopilot/locks/heading-true-hold", 0);
		setprop("/autopilot/locks/speed", "wind");
	}
}

var toggleSpeedZero = func()
{
	var aLock = getprop("/autopilot/locks/speed");
	if (aLock == "speed-with-throttle")
	{
		setprop("/fdm/jsbsim/crew/helmsman/enabled" , 1);
		setprop("/autopilot/settings/target-speed-kt" , 0);
		setprop("/autopilot/settings/target-speed-kmh", 0);
		setprop("/autopilot/locks/speed", "");
		SetEngineState(0, 1);
		SetEngineState(1, 1);
		SetEngineState(2, 1);
		SetEngineState(3, 1);
		SetEngineState(4, 1);
		SetEngineState(5, 1);
	}
	else if (aLock == "wind")
	{
		SetEngineState(0, 0);
		SetEngineState(1, 1);
		SetEngineState(2, 1);
		SetEngineState(3, 1);
		SetEngineState(4, 1);
		SetEngineState(5, 0);
		setprop("/autopilot/settings/target-speed-kt" , 0);
		setprop("/autopilot/settings/target-speed-kmh", 0);
		setprop("/fdm/jsbsim/crew/helmsman/enabled" , 1);
	#	setprop("/autopilot/locks/heading", "dg-heading-hold");
		setprop("/autopilot/locks/speed", "speed-with-throttle");
	}
	else
	{
		SetEngineState(0, 1);
		SetEngineState(1, 1);
		SetEngineState(2, 1);
		SetEngineState(3, 1);
		SetEngineState(4, 1);
		SetEngineState(5, 1);
		setprop("/autopilot/settings/target-speed-kt" , 0);
		setprop("/autopilot/settings/target-speed-kmh", 0);
		setprop("/fdm/jsbsim/crew/helmsman/enabled" , 1);
	#	setprop("/autopilot/locks/heading", "dg-heading-hold");
		setprop("/autopilot/locks/speed", "speed-with-throttle");
	}
}

var AutoLand_timer = 0;

var autoLand = func()
{
	if (getprop("/autopilot/settings/autoland") == 1)
	{
		setprop("/autopilot/settings/autoland", 0);
		AutoLand_timer.stop();
		setprop("/autopilot/locks/speed", "");
		gui.popupTip("Automatic landing aborted");
		SetEngineState(1, 1);
		SetEngineState(2, 1);
		SetEngineState(3, 1);
		SetEngineState(4, 1);
		return;
	}
	setprop("/autopilot/settings/autoland", 1);
	SetEngineState(0, 0);
	SetEngineState(1, 1);
	SetEngineState(2, 1);
	SetEngineState(3, 1);
	SetEngineState(4, 1);
	SetEngineState(5, 0);
	setprop("/autopilot/settings/target-speed-kt" , 0);
	setprop("/autopilot/settings/target-speed-kmh", 0);
	setprop("/fdm/jsbsim/crew/helmsman/enabled" , 1);
	setprop("/autopilot/locks/speed", "speed-with-throttle");
	gui.popupTip("Slowing down to windspeed");
	AutoLand_timer = maketimer(1, func {
		if (getprop("/instrumentation/gps/indicated-ground-speed-kt") < getprop("/environment/wind-speed-kt") + 2)
		{
			AutoLand_timer.stop();
			if (getprop("/environment/wind-speed-kt") < 2.5)
			{
				gui.popupTip("Slowing down to zero");
				setprop("/autopilot/settings/target-speed-kt" , 0);
				setprop("/autopilot/settings/target-speed-kmh", 0);
				setprop("/fdm/jsbsim/crew/helmsman/enabled" , 1);
				setprop("/autopilot/locks/heading-enable", 1);
				setprop("/autopilot/locks/heading-dg-hold", 1);
				setprop("/autopilot/locks/heading-true-hold", 0);
				setprop("/autopilot/locks/speed", "speed-with-throttle");
				AutoLand_timer = maketimer(0.5, func {
					if (getprop("/instrumentation/gps/indicated-ground-speed-kt") < 2.0)
					{
						AutoLand_timer.stop();
						gui.popupTip("Prepare for final landing");
						ground_crew.place_ground_crew(geo.aircraft_position(), getprop("/orientation/heading-deg"));
						if (getprop("/position/altitude-agl-ft") < 550)
						{
							setprop("/autopilot/locks/speed", "");
							ground_crew.activate();
							disEngageClutch();
							SetEngineState(0, 0);
							SetEngineState(1, 0);
							SetEngineState(2, 0);
							SetEngineState(3, 0);
							SetEngineState(4, 0);
							SetEngineState(5, 0);
							setprop("/autopilot/settings/autoland", 0);
						}
					}
				});
			}
			else
			{
				gui.popupTip("Turning into the wind");
				setprop("/autopilot/locks/speed", "wind");
				var oldHead = getprop("/orientation/heading-deg");
				AutoLand_timer = maketimer(0.5, func {
					var currentHead = getprop("/orientation/heading-deg");
					var currentWind = getprop("/environment/wind-from-heading-deg");
					if ((math.abs(oldHead - currentHead) < 0.07) and ((currentHead > geo.normdeg(currentWind - 1)) and (currentHead < geo.normdeg(currentWind + 1))))
					{
						AutoLand_timer.stop();
						gui.popupTip("Slowing down to zero");
						setprop("/autopilot/settings/target-speed-kt" , 0);
						setprop("/autopilot/settings/target-speed-kmh", 0);
						setprop("/fdm/jsbsim/crew/helmsman/enabled" , 1);
						setprop("/autopilot/locks/heading-enable", 1);
						setprop("/autopilot/locks/heading-dg-hold", 1);
						setprop("/autopilot/locks/heading-true-hold", 0);
						setprop("/autopilot/locks/speed", "speed-with-throttle");
						AutoLand_timer = maketimer(0.5, func {
							if (getprop("/instrumentation/gps/indicated-ground-speed-kt") < 2.0)
							{
								AutoLand_timer.stop();
								gui.popupTip("Prepare for final landing");
								ground_crew.place_ground_crew(geo.aircraft_position(), getprop("/orientation/heading-deg"));
								if (getprop("/position/altitude-agl-ft") < 550)
								{
									setprop("/autopilot/locks/speed", "");
									ground_crew.activate();
									disEngageClutch();
									SetEngineState(0, 0);
									SetEngineState(1, 0);
									SetEngineState(2, 0);
									SetEngineState(3, 0);
									SetEngineState(4, 0);
									SetEngineState(5, 0);
									setprop("/autopilot/settings/autoland", 0);
								}
							}
						});
						AutoLand_timer.start();
					}
					oldHead = currentHead;
				});
			}
			AutoLand_timer.start();
		}
	});
	AutoLand_timer.start();
}

var reinitWeights = func()
{
#	gui.popupTip("Starting reloading Bombs");
	setprop("/sim/messages/ground", "Starting reloading Bombs");
	setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[4]", 66225.0);
	setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[5]", 66225.0);
	setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[0]", QUICK_RELEASE_BAG_CAPACITY);
	setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[1]", QUICK_RELEASE_BAG_CAPACITY);
	setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[2]", QUICK_RELEASE_BAG_CAPACITY);
	setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[3]", QUICK_RELEASE_BAG_CAPACITY);
	setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[6]", QUICK_RELEASE_BAG_CAPACITY);
	setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[7]", QUICK_RELEASE_BAG_CAPACITY);
	setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[8]", QUICK_RELEASE_BAG_CAPACITY);
	setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[9]", QUICK_RELEASE_BAG_CAPACITY);
	setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[10]", QUICK_RELEASE_BAG_CAPACITY);
	setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[11]", QUICK_RELEASE_BAG_CAPACITY);
	if (getprop("/payload/weight[0]/selected") == "Bombs 100 kg")
	{
		for (var i = 0; i < 8; i = i + 1)
		{
			setprop("/ai/submodels/submodel[" ~ i ~ "]/count", 1);
			setprop("/sim/weapons/bomb[" ~ i ~ "]/dropped", 0);
			setprop("/sim/weapons/bomb[" ~ i ~ "]/present", 1);
			setprop("/sim/weapons/bomb[" ~ i ~ "]/tobedropped", 0);
		}
		setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[23]", 1763.7);
	}
	if (getprop("/payload/weight[1]/selected") == "Bombs 100 kg")
	{
		for (var i = 8; i < 16; i = i + 1)
		{
			setprop("/ai/submodels/submodel[" ~ i ~ "]/count", 1);
			setprop("/sim/weapons/bomb[" ~ i ~ "]/dropped", 0);
			setprop("/sim/weapons/bomb[" ~ i ~ "]/present", 1);
			setprop("/sim/weapons/bomb[" ~ i ~ "]/tobedropped", 0);
		}
		setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[24]", 1763.7);
	}
	if (getprop("/payload/weight[2]/selected") == "Bombs 100 kg")
	{
		for (var i = 16; i < 24; i = i + 1)
		{
			setprop("/ai/submodels/submodel[" ~ i ~ "]/count", 1);
			setprop("/sim/weapons/bomb[" ~ i ~ "]/dropped", 0);
			setprop("/sim/weapons/bomb[" ~ i ~ "]/present", 1);
			setprop("/sim/weapons/bomb[" ~ i ~ "]/tobedropped", 0);
		}
		setprop("/fdm/jsbsim/inertia/pointmass-weight-lbs[25]", 1763.7);
	}
	setprop("/sim/weapons/dropping", 0);
	SetEngineState(0, 0);
	SetEngineState(1, 0);
	SetEngineState(2, 0);
	SetEngineState(3, 0);
	SetEngineState(4, 0);
	SetEngineState(5, 0);
	# repair propellers
	setprop("/controls/engines/engine[0]/defect", 1);
	setprop("/controls/engines/engine[1]/defect", 1);
	setprop("/controls/engines/engine[2]/defect", 1);
	setprop("/controls/engines/engine[3]/defect", 1);
	setprop("/controls/engines/engine[4]/defect", 1);
	setprop("/controls/engines/engine[5]/defect", 1);
#	gui.popupTip("Starting refilling Fuel");
	setprop("/sim/messages/ground", "Starting refilling Fuel");
	for (var t = 0; t < 14; t = t + 1)
	{
		setprop("/consumables/fuel/tank[" ~ t ~ "]/level-norm", 1);
		setprop("/consumables/fuel/tank[" ~ t ~ "]/empty", 0);
	}
	setprop("/sim/rendering/redout/oxygen-count", 28800);
	setprop("/controls/flight/trim", 0);
	setprop("/fdm/jsbsim/inertia/pointmass-location-X-inches[16]", baseTrimPos);
	setprop("/controls/flight/rudder", 0);
	setprop("/controls/flight/elevator", 0);
	setprop("/controls/gas/pressurecontrol", 0);
	setprop("/sim/messages/ground", "Starting refilling Hydrogen gas");
	setprop("/autopilot/settings/vertical-speed-fps" , 0);
	setprop("/controls/gas/valve-cmd-norm[0]", 0);
	setprop("/controls/gas/valve-cmd-norm[1]", 0);
	setprop("/controls/gas/valve-cmd-norm[2]", 0);
	setprop("/controls/gas/valve-cmd-norm[3]", 0);
	setprop("/controls/gas/valve-cmd-norm[4]", 0);
	setprop("/controls/gas/valve-cmd-norm[5]", 0);
	setprop("/controls/gas/valve-cmd-norm[6]", 0);
	setprop("/controls/gas/valve-cmd-norm[7]", 0);
	setprop("/controls/gas/valve-cmd-norm[8]", 0);
	setprop("/controls/gas/valve-cmd-norm[9]", 0);
	setprop("/controls/gas/valve-cmd-norm[10]", 0);
	setprop("/controls/gas/valve-cmd-norm[11]", 0);
	setprop("/controls/gas/valve-cmd-norm[12]", 0);
	setprop("/controls/gas/valve-cmd-norm[13]", 0);
	setprop("/controls/ballast/action[0]", 0);
	setprop("/controls/ballast/action[1]", 0);
	setprop("/controls/ballast/action[2]", 0);
	setprop("/controls/ballast/action[3]", 0);
	setprop("/controls/ballast/action[4]", 0);
	setprop("/controls/ballast/action[5]", 0);
	setprop("/controls/ballast/action[6]", 0);
	setprop("/controls/ballast/action[7]", 0);
	setprop("/controls/ballast/action[8]", 0);
	setprop("/controls/ballast/action[9]", 0);
	setprop("/controls/ballast/current_fore_ballast_index", 0);
	setprop("/controls/ballast/current_aft_ballast_index", 0);
	setprop("/controls/ballast/release[0]", 0);
	setprop("/controls/ballast/release[1]", 0);
	setprop("/controls/ballast/release[2]", 0);
	setprop("/controls/ballast/release[3]", 0);
	setprop("/controls/ballast/release[4]", 0);
	setprop("/controls/ballast/release[5]", 0);
	setprop("/controls/ballast/release[6]", 0);
	setprop("/controls/ballast/release[7]", 0);
	setprop("/controls/ballast/release[8]", 0);
	setprop("/controls/ballast/release[9]", 0);
	setprop("controls/gas/destroypressure[0]", 0);
	setprop("controls/gas/destroypressure[1]", 0);
	setprop("controls/gas/destroypressure[2]", 0);
	setprop("controls/gas/destroypressure[3]", 0);
	setprop("controls/gas/destroypressure[4]", 0);
	setprop("controls/gas/destroypressure[5]", 0);
	setprop("controls/gas/destroypressure[6]", 0);
	setprop("controls/gas/destroypressure[7]", 0);
	setprop("controls/gas/destroypressure[8]", 0);
	setprop("controls/gas/destroypressure[9]", 0);
	setprop("controls/gas/destroypressure[10]", 0);
	setprop("controls/gas/destroypressure[11]", 0);
	setprop("controls/gas/destroypressure[12]", 0);
	setprop("controls/gas/destroypressure[13]", 0);
	var Pressure = getprop("/fdm/jsbsim/atmosphere/P-psf");
	var fullness = getprop("/controls/gas/maxfullness") / 100.0;
	for (var g = 0; g < 14; g = g + 1)
	{
		var Temperature = getprop("/fdm/jsbsim/buoyant_forces/gas-cell[" ~ g ~ "]/temp-R");
		var Volume = getprop("/fdm/jsbsim/buoyant_forces/gas-cell[" ~ g ~ "]/max_volume-ft3");
		var Contents = Pressure * Volume * fullness / (3.4071 * Temperature);
		setprop("/fdm/jsbsim/buoyant_forces/gas-cell[" ~ g ~ "]/contents-mol", Contents);
	}
	L70.CLIP.reinit();
	settimer(func {
		setprop("/controls/gas/pressurecontrol", 1);
		setprop("/sim/messages/ground", "Ship is ready again!");
	}, 10);
}

var adjustRudder = func(ap, rudder)
{
	if (getprop("/autopilot/locks/heading-enable") and getprop("/autopilot/locks/heading-dg-hold"))
	{
		var value = getprop("/fdm/jsbsim/crew/helmsman/heading-magnetic-cmd-deg") + ap;
		if (value > 360)
		{
			value = value - 360;
		}
		if (value < 0)
		{
			value = value + 360;
		}
		setprop("/autopilot/settings/true-heading-deg", value);
		setprop("/fdm/jsbsim/crew/helmsman/heading-magnetic-cmd-deg", value);
	}
	else
	{
		if (getprop("/autopilot/locks/heading-enable") and getprop("/autopilot/locks/heading-true-hold"))
		{
			var value = getprop("/autopilot/settings/true-heading-deg") + ap;
			if (value > 360)
			{
				value = value - 360;
			}
			if (value < 0)
			{
				value = value + 360;
			}
			setprop("/autopilot/settings/true-heading-deg", value);
		}
		else
		{
			var value = getprop("/controls/flight/rudder") + rudder;
			if (value < -1.0)
			{
				value = -1.0;
			}
			if (value > 1.0)
			{
				value = 1.0;
			}
			setprop("/controls/flight/rudder", value);
		}
	}
}

var adjustElevator = func(ap1, ap2, elevator)
{
	if (getprop("/autopilot/locks/altitude-enable"))
	{
		if (getprop("/autopilot/locks/altitude-pitch") or getprop("/autopilot/locks/altitude-speed"))
		{
			var value = getprop("/autopilot/settings/target-pitch-deg") + ap1;
			if (value > 20)
			{
				value = 20;
			}
			if (value < -20)
			{
				value = -20;
			}
			setprop("/autopilot/settings/target-pitch-deg", value);
		}
		else if (getprop("/autopilot/locks/altitude-level"))
		{
			setprop("/autopilot/settings/target-altitude-ft", getprop("/autopilot/settings/target-altitude-ft") + (ap2 / 0.3048));
			setprop("/autopilot/settings/target-altitude-m", getprop("/autopilot/settings/target-altitude-ft") * 0.3048);
		}
	}
	else
	{
		setprop("/controls/flight/elevator", getprop("/controls/flight/elevator") + elevator);
	}
}

################################################################################
#put_remove_model places a new model at the location specified and then removes
# it time_sec later 
#it puts out 12 models/sec so normally time_sec=.4 or thereabouts it plenty of time to let it run
# If time_sec is too short then no particles will be emitted.  Typical problem is 
# many rounds from a gun slow FG's framerate to a crawl just as it is time to emit the 
# particles.  If time_sec is slower than the frame length then you get zero particle.
# Smallest safe value for time_sec is maybe .3 .4 or .5 seconds.
# 
var put_remove_model = func(lat_deg=nil, lon_deg=nil, elev_m=nil, time_sec=nil, startSize_m=nil, endSize_m=1, path="Models/bombable/Fire-Particles/flack-impact.xml" ) {

  if (lat_deg==nil or lon_deg==nil or elev_m==nil) { return; } 
 
  var delay_sec=0.1; #particles/models seem to cause FG crash *sometimes* when appearing within a model
  #we try to reduce this by making the smoke appear a fraction of a second later, after
  # the a/c model has moved out of the way. (possibly moved, anyway--depending on it's speed)  

 # debprint ("Bombable: Placing flack");
         
  settimer ( func {
    #start & end size in particle system appear to be in feet
    if (startSize_m!=nil) setprop ("/bombable/fire-particles/flack-startsize", startSize_m);
    if (endSize_m!=nil) setprop ("/bombable/fire-particles/flack-endsize", endSize_m);

    fgcommand("add-model", flackNode=props.Node.new({ 
            "path": path, 
            "latitude-deg": lat_deg, 
            "longitude-deg":lon_deg, 
            "elevation-ft": elev_m / 0.3048,
            "heading-deg"  : 0,
            "pitch-deg"    : 0,
            "roll-deg"     : 0, 
            "enable-hot"   : 0,
  
            
              
    }));
    
  var flackModelNodeName= flackNode.getNode("property").getValue();
  
  #add the -prop property in /models/model[X] for each of lat, long, elev, etc
  foreach (name; ["latitude-deg","longitude-deg","elevation-ft", "heading-deg", "pitch-deg", "roll-deg"]){
   setprop(  flackModelNodeName ~"/"~ name ~ "-prop",flackModelNodeName ~ "/" ~ name );
  }  
  
#  debprint ("Bombable: Placed flack, ", flackModelNodeName);
  
  settimer ( func { props.globals.getNode(flackModelNodeName).remove();}, time_sec);

  }, delay_sec);   

}

##############################################################
#Start a fire on terrain, size depending on ballisticMass_lb
#location at lat/lon
#
var start_terrain_fire = func ( lat_deg, lon_deg, alt_m=0, ballisticMass_lb=1.2 ) {

  var info = geodinfo(lat_deg, lon_deg);
  var smokeEndsize = rand()*75+33;
  setprop ("/bombable/fire-particles/smoke-endsize-small", smokeEndsize);
  var smokeEndsize = rand()*125+60;
  setprop ("/bombable/fire-particles/smoke-endsize-large", smokeEndsize);
  var smokeStartsize=rand()*10 + 5;
  setprop ("/bombable/fire-particles/smoke-startsize", smokeStartsize);
  setprop ("/bombable/fire-particles/smoke-startsize-small", smokeStartsize * (rand()/2 + 0.5));
  setprop ("/bombable/fire-particles/smoke-startsize-very-small", smokeStartsize * (rand()/8 + 0.2));
  setprop ("/bombable/fire-particles/smoke-startsize-large", smokeStartsize* (rand()*4 + 1));
  #get the altitude of the terrain
  if (info != nil)
  {
      #if it's water we don't set a fire . . . TODO make a different explosion or fire effect for water 
      if (typeof(info[1])=="hash" and contains(info[1], "solid") and info[1].solid==0) return;
     # else debprint (info);
      
      #we go with explosion point if possible, otherwise the height of terrain at this point
      if (alt_m==nil) alt_m=info[0];
      if (alt_m==nil) alt_m=0; 
      
  }
  time_sec=75; 
  fp1="Aircraft/L70/Models/bombable/Fire-Particles/fire-particles-large.xml";
  put_remove_model(lat_deg:lat_deg, lon_deg:lon_deg, elev_m:alt_m, time_sec:time_sec, startSize_m: nil, endSize_m:nil, path:fp1 );
  time_sec1=300; 
  fp="Aircraft/L70/Models/bombable/Fire-Particles/smoke-particles-large.xml";
  put_remove_model(lat_deg:lat_deg, lon_deg:lon_deg, elev_m:alt_m, time_sec:time_sec1, startSize_m: nil, endSize_m:nil, path:fp );
  ##put it out, but slowly, for large impacts
  if (ballisticMass_lb>50)
  {
    time_sec2=75; fp2="Aircraft/L70/Models/bombable/Fire-Particles/fire-particles-very-small.xml";  
    settimer (func { put_remove_model(lat_deg:lat_deg, lon_deg:lon_deg, elev_m:alt_m, time_sec:time_sec2, startSize_m: nil, endSize_m:nil, path:fp2 )} , time_sec);
    
    time_sec3=75; fp3="Aircraft/L70/Models/bombable/Fire-Particles/fire-particles-very-very-small.xml";  
    settimer (func { put_remove_model(lat_deg:lat_deg, lon_deg:lon_deg, elev_m:alt_m, time_sec:time_sec3, startSize_m: nil, endSize_m:nil, path:fp3 )} , time_sec+time_sec2);
    
    time_sec4=75; fp4="Aircraft/L70/Models/bombable/Fire-Particles/fire-particles-very-very-very-small.xml";  
    settimer (func { put_remove_model(lat_deg:lat_deg, lon_deg:lon_deg, elev_m:alt_m, time_sec:time_sec4, startSize_m: nil, endSize_m:nil, path:fp4 )} , time_sec+time_sec2+time_sec3);
  
  }

}
###############################################################################
# Utility functions

# Set up aTransfer fluid between two properties without losing or creating any.
#  amount [lb]
#  rate   [lb/sec]
var SmoothTransfer = {
    # Set up a smooth value transfer between two properties.
    #  from, to       : property paths or property nodes
    #  rate           : units/sec. MUST be positive.
    #  amount         : the amount to transfer. MUST be positive.
    #  stop_at_zero   : stop if the from property reaches 0.0.
    new: func(from, to, rate, amount=-1, stop_at_zero=1) {
        var m = { parents: [SmoothTransfer] };
        m.from         = aircraft.makeNode(from);
        m.to           = aircraft.makeNode(to);
        m.left         = amount;
        m.rate         = rate;
        m.stop_at_zero = stop_at_zero;
        m.last_time    = getprop("/sim/time/elapsed-sec");
        settimer(func{ m._loop_(); }, 0.0);
        return m;
    },
    stop: func {
        me.left = 0.0;
    },
    _loop_: func()
    {
        var t = getprop("/sim/time/elapsed-sec");
        var a = me.rate * (t - me.last_time);
        a = (a < me.left) ? a : me.left;
        if (me.stop_at_zero)
        {
            var f = me.from.getValue();
            a = (a < f) ? a : f;
        }
        me.from.setValue(me.from.getValue() - a);
        me.to.setValue(me.to.getValue() + a);
        me.left -= a;
        me.last_time = t;
        if (me.left > 0.0)
        {
            settimer(func{ me._loop_(); }, 0.0);
        }
    }
};

###########################################################################
## MP integration of user's fixed local mooring locations.
## NOTE: Should this be here or elsewhere?
#settimer(func { mp_network_init(1); }, 0.1);

###############################################################################
# About dialog.

var ABOUT_DLG = 0;

var dialog = {
#################################################################
    init : func (x = nil, y = nil) {
        me.x = x;
        me.y = y;
        me.bg = [0, 0, 0, 0.3];    # background color
        me.fg = [[1.0, 1.0, 1.0, 1.0]]; 
        #
        # "private"
        me.title = "About L70";
        me.dialog = nil;
        me.namenode = props.Node.new({"dialog-name" : me.title });
    },
#################################################################
    create : func {
        if (me.dialog != nil)
            me.close();

        me.dialog = gui.Widget.new();
        me.dialog.set("name", me.title);
        if (me.x != nil)
            me.dialog.set("x", me.x);
        if (me.y != nil)
            me.dialog.set("y", me.y);

        me.dialog.set("layout", "vbox");
        me.dialog.set("default-padding", 0);

        var titlebar = me.dialog.addChild("group");
        titlebar.set("layout", "hbox");
        titlebar.addChild("empty1").set("stretch", 1);
        titlebar.addChild("text").set("label", "About L70");
        titlebar.addChild("empty2").set("stretch", 1);
        var w = titlebar.addChild("button");
        w.set("pref-width", 16);
        w.set("pref-height", 16);
        w.set("legend", "");
        w.set("default", 0);
        w.set("key", "esc");
        w.setBinding("nasal", "L70.dialog.destroy(); ");
        w.setBinding("dialog-close");
        me.dialog.addChild("hrule");

        var content = me.dialog.addChild("group");
        content.set("layout", "vbox");
        content.set("halign", "center");
        content.set("default-padding", 5);
        props.globals.initNode("sim/about/text",
             "Zeppelin L70 (LZ 112) airship for FlightGear\n" ~
             "Copyright (C) 2020 - 2023  Franz Schmid\n\n" ~
             "Copyright (C) 2010 - 2016  Anders Gidenstam\n\n" ~
             "FlightGear flight simulator\n" ~
             "Copyright (C) 1996 - 2023  http://www.flightgear.org\n\n" ~
             "This is free software, and you are welcome to\n" ~
             "redistribute it under certain conditions.\n" ~
             "See the GNU GENERAL PUBLIC LICENSE Version 2 for the details.",
             "STRING");
        var text = content.addChild("textbox");
        text.set("halign", "fill");
        #text.set("slider", 20);
        text.set("pref-width", 400);
        text.set("pref-height", 300);
        text.set("editable", 0);
        text.set("property", "sim/about/text");

        #me.dialog.addChild("hrule");

        fgcommand("dialog-new", me.dialog.prop());
        fgcommand("dialog-show", me.namenode);
    },
#################################################################
    close : func {
        fgcommand("dialog-close", me.namenode);
    },
#################################################################
    destroy : func {
        ABOUT_DLG = 0;
        me.close();
        delete(gui.dialog, "\"" ~ me.title ~ "\"");
    },
#################################################################
    show : func {
        if (!ABOUT_DLG) {
            ABOUT_DLG = 1;
            me.init(400, getprop("/sim/startup/ysize") - 500);
            me.create();
        }
    }
};
###############################################################################

# Popup the about dialog.
var about = func {
    dialog.show();
}
###############################################################################
## Ground party.

var ground_crew = {
    ##################################################
    init : func {
        me.UPDATE_INTERVAL = 0.42;
        me.loopid = 0;
        # There are two handling guy parties.
        me.position = geo.aircraft_position();
        me.pos = [geo.aircraft_position(), geo.aircraft_position(), geo.aircraft_position(), geo.aircraft_position()];
        me.connected = [props.globals.getNode("/fdm/jsbsim/landing-party/wire-connected[0]"),
                        props.globals.getNode("/fdm/jsbsim/landing-party/wire-connected[1]"),
                        props.globals.getNode("/fdm/jsbsim/landing-party/wire-connected[2]"),
						props.globals.getNode("/fdm/jsbsim/landing-party/wire-connected[3]")];
        me.wire_length = props.globals.getNode("/fdm/jsbsim/landing-party/wire-length-ft");
        me.wire_length2 = props.globals.getNode("/fdm/jsbsim/landing-party/wire2-length-ft");
        me.model = {local : [nil, nil, nil, nil]};
        me.wind_heading = props.globals.getNode("/environment/wind-from-heading-deg");
        me.active = 0;
        me.active = 1;
        me.connected[0].setValue(0.99);
        me.connected[1].setValue(0.99);
        me.connected[2].setValue(0.99);
        me.connected[3].setValue(0.99);
        me.reset();
        print("Ground crew ... Standing by.");
    },
    ##################################################
    # Place the ground crew.
    place_ground_crew : func (pos, heading=nil, altitude=nil, name="local")
	{
        if (heading == nil) {
            me.heading = me.wind_heading.getValue();
        } else {
            me.heading = heading;
        }
        me.position = pos;
        me.pos[0].set(pos);
        me.pos[1].set(pos);
        me.pos[2].set(pos);
        me.pos[3].set(pos);
        me.pos[0].apply_course_distance(me.heading - 45.0, 20.0);
        me.pos[1].apply_course_distance(me.heading + 45.0, 20.0);
        me.pos[2].apply_course_distance(me.heading, -211.0);
        me.pos[3].apply_course_distance(me.heading, -211.0);
        me.pos[2].apply_course_distance(me.heading - 135.0, -20.0);
        me.pos[3].apply_course_distance(me.heading + 135.0, -20.0);
        if (altitude == nil)
		{
            me.pos[0].set_alt(geodinfo(me.pos[0].lat(), me.pos[0].lon())[0]);
            me.pos[1].set_alt(geodinfo(me.pos[1].lat(), me.pos[1].lon())[0]);
            me.pos[2].set_alt(geodinfo(me.pos[2].lat(), me.pos[2].lon())[0]);
            me.pos[3].set_alt(geodinfo(me.pos[3].lat(), me.pos[3].lon())[0]);
        }
        else
		{
            me.pos[0].set_alt(altitude);
            me.pos[1].set_alt(altitude);
            me.pos[2].set_alt(altitude);
            me.pos[3].set_alt(altitude);
        }  
    #    print("ground_crew: Handling parties at ");
    #    me.pos[0].dump();
	#	me.pos[1].dump();
	#	me.pos[2].dump();
	#	me.pos[3].dump();

        setprop("/fdm/jsbsim/landing-party/latitude-deg[0]", me.pos[0].lat());
        setprop("/fdm/jsbsim/landing-party/longitude-deg[0]", me.pos[0].lon());
        setprop("/fdm/jsbsim/landing-party/altitude-ft[0]",  me.pos[0].alt() * M2FT);
        setprop("/fdm/jsbsim/landing-party/latitude-deg[1]", me.pos[1].lat());
        setprop("/fdm/jsbsim/landing-party/longitude-deg[1]", me.pos[1].lon());
        setprop("/fdm/jsbsim/landing-party/altitude-ft[1]", me.pos[1].alt() * M2FT);
        setprop("/fdm/jsbsim/landing-party/latitude-deg[2]", me.pos[2].lat());
        setprop("/fdm/jsbsim/landing-party/longitude-deg[2]", me.pos[2].lon());
        setprop("/fdm/jsbsim/landing-party/altitude-ft[2]", me.pos[2].alt() * M2FT);
        setprop("/fdm/jsbsim/landing-party/latitude-deg[3]", me.pos[3].lat());
        setprop("/fdm/jsbsim/landing-party/longitude-deg[3]", me.pos[3].lon());
        setprop("/fdm/jsbsim/landing-party/altitude-ft[3]", me.pos[3].alt() * M2FT);
		
		var backPos = geo.aircraft_position();
        backPos.apply_course_distance(getprop("/orientation/heading-deg"), -211.0);
		setprop("/fdm/jsbsim/landing-party/latitude-degB", backPos.lat());
		setprop("/fdm/jsbsim/landing-party/longitude-degB", backPos.lon());
    
        if (me.model.local[0] != nil) me.model.local[0].remove();
        if (me.model.local[1] != nil) me.model.local[1].remove();
        if (me.model.local[2] != nil) me.model.local[2].remove();
        if (me.model.local[3] != nil) me.model.local[3].remove();
        me.model.local[0] = geo.put_model(getprop("/sim/aircraft-dir") ~ "/Models/GroundCrew/wire-party.xml", me.pos[0], me.heading + 135.0);
        me.model.local[1] = geo.put_model(getprop("/sim/aircraft-dir") ~ "/Models/GroundCrew/wire-party.xml", me.pos[1], me.heading - 135.0);
        me.model.local[2] = geo.put_model(getprop("/sim/aircraft-dir") ~ "/Models/GroundCrew/wire-party.xml", me.pos[2], me.heading + 45.0);
        me.model.local[3] = geo.put_model(getprop("/sim/aircraft-dir") ~ "/Models/GroundCrew/wire-party.xml", me.pos[3], me.heading - 45.0);
        broadcast.send(message_id["place_ground_crew"] ~
                       Binary.encodeCoord(me.pos[0]) ~
                       Binary.encodeCoord(me.pos[1]) ~
                       Binary.encodeCoord(me.pos[2]) ~
                       Binary.encodeCoord(me.pos[3]) ~
                       Binary.encodeDouble(me.heading));
    },
    ##################################################
    place_remote_ground_crew : func (key, pos1, pos2, pos3, pos4, heading) {
        if (!contains(me.model, key)) me.model[key] = [nil, nil];

        if (me.model[key][0] != nil) me.model[key][0].remove();
        if (me.model[key][1] != nil) me.model[key][1].remove();
        if (me.model[key][2] != nil) me.model[key][2].remove();
        if (me.model[key][3] != nil) me.model[key][3].remove();
        me.model[key][0] = geo.put_model(getprop("/sim/aircraft-dir") ~ "/Models/GroundCrew/wire-party.xml", pos1, heading + 135.0);
        me.model[key][1] = geo.put_model(getprop("/sim/aircraft-dir") ~ "/Models/GroundCrew/wire-party.xml", pos2, heading - 135.0);
        me.model[key][2] = geo.put_model(getprop("/sim/aircraft-dir") ~ "/Models/GroundCrew/wire-party.xml", pos3, heading + 45.0);
        me.model[key][3] = geo.put_model(getprop("/sim/aircraft-dir") ~ "/Models/GroundCrew/wire-party.xml", pos4, heading - 45.0);
    },
    ##################################################
    remove_remote_ground_crew : func (key) {
        if (!contains(me.model, key)) return;
        if (me.model[key][0] != nil) me.model[key][0].remove();
        if (me.model[key][1] != nil) me.model[key][1].remove();
        if (me.model[key][2] != nil) me.model[key][2].remove();
        if (me.model[key][3] != nil) me.model[key][3].remove();
    },
    ##################################################
    let_go : func {
        if (me.connected[0].getValue() or me.connected[1].getValue())
            me.announce("Handling guys released!");
        me.active = 0;
        me.connected[0].setValue(0.0);
        me.connected[1].setValue(0.0);
        me.connected[2].setValue(0.0);
        me.connected[3].setValue(0.0);
        me.wire_length.setValue(650.0);
    },
    ##################################################
    activate : func {
        me.active = 1;
        me.wire_length.setValue(650.0);
        me.wire_length2.setValue(650.0);
     #   me.place_ground_crew(me.position);
        me.announce("Ready for landing!");
    },
    ##################################################
    announce : func(msg) {
        setprop("/sim/messages/ground", msg);
    },
    ##################################################
    update : func {
        if (!me.active) return;
        
        if ((me.connected[0].getValue() < 1.0) and (getprop("/fdm/jsbsim/landing-party/total-distance-ft[0]") < 2.0*getprop("/fdm/jsbsim/landing-party/wire-length-ft")))
		{
            me.connected[0].setValue(1.0);
            me.announce("Left fore handling guy secured!");
        }
        if ((me.connected[1].getValue() < 1.0) and (getprop("/fdm/jsbsim/landing-party/total-distance-ft[1]") < 2.0*getprop("/fdm/jsbsim/landing-party/wire-length-ft")))
		{
            me.connected[1].setValue(1.0);
            me.announce("Right fore handling guy secured!");
        }
        if ((me.connected[2].getValue() < 1.0) and (getprop("/fdm/jsbsim/landing-party/total-distance-ft[2]") < 2.0*getprop("/fdm/jsbsim/landing-party/wire2-length-ft")))
		{
            me.connected[2].setValue(1.0);
            me.announce("Left aft handling guy secured!");
        }
        if ((me.connected[3].getValue() < 1.0) and (getprop("/fdm/jsbsim/landing-party/total-distance-ft[3]") < 2.0*getprop("/fdm/jsbsim/landing-party/wire2-length-ft")))
		{
            me.connected[3].setValue(1.0);
            me.announce("Right aft handling guy secured!");
        }
        if ((me.connected[0].getValue() >= 0.99) and (me.connected[1].getValue() >= 0.99) and (me.wire_length.getValue() == 650.0))
		{
            interpolate(me.wire_length, 40.0, 30.0);
        }
        if ((me.connected[2].getValue() >= 0.99) and (me.connected[3].getValue() >= 0.99) and (me.wire_length2.getValue() == 650.0))
		{
            interpolate(me.wire_length2, 40.0, 30.0);
        }
		var backPos = geo.aircraft_position();
        backPos.apply_course_distance(getprop("/orientation/heading-deg"), -211.0);
		setprop("/fdm/jsbsim/landing-party/latitude-degB", backPos.lat());
		setprop("/fdm/jsbsim/landing-party/longitude-degB", backPos.lon());
    },
    ##################################################
    reset : func {
        me.loopid += 1;
        me._loop_(me.loopid);
    },
    ##################################################
    _loop_ : func(id) {
        id == me.loopid or return;
        me.update();
        settimer(func { me._loop_(id); }, me.UPDATE_INTERVAL);
    }
};

var helpEngControl = func()
{
	var level = getprop("controls/engines/helpengine-lever");
	var mains = getprop("systems/electrical/mains");
	if (getprop("/fdm/jsbsim/propulsion/total-fuel-lbs") > 0)
	{
		if (level <= 0.5)
		{
			level = 1.0;
			if (mains < 1)
			{
				interpolate("systems/electrical/mains", 1, 2.0);
			}
		}
		else
		{
			level = 0.0;
			if (mains < 1.5)
			{
				interpolate("systems/electrical/mains", 0, 2.0);
			}
		}
	}
	else
	{
		level = 0.0;
		if (mains < 1.5)
		{
			interpolate("systems/electrical/mains", 0, 2.0);
		}
	}
	interpolate("controls/engines/helpengine-lever", level, 2.0);
}

var weather_dialog = gui.Dialog.new("/sim/gui/dialogs/L70/weather/dialog", "Aircraft/L70/Dialogs/weather-report.xml");
var autopilot_dialog = gui.Dialog.new("sim/gui/dialogs/autopilot/dialog", "Aircraft/L70/Dialogs/autopilot.xml");
var aircraft_dialog = gui.Dialog.new("/sim/gui/dialogs/aircraft/dialog", "Aircraft/L70/Dialogs/aircraft-dialog.xml");
var route_dialog = gui.Dialog.new("/sim/gui/dialogs/L70/route/dialog", "Aircraft/L70/Dialogs/create-route.xml");
var config_dialog = gui.Dialog.new("/sim/gui/dialogs/L70/payload/dialog", "Aircraft/L70/Dialogs/configure-payload.xml");
var pushback_dialog = gui.Dialog.new("/sim/gui/dialogs/L70/pushback/dialog", "Aircraft/L70/Dialogs/pushback.xml");
var view_select_dialog = gui.Dialog.new("/sim/gui/dialogs/L70/views/dialog", "Aircraft/L70/Dialogs/view-select-dialog.xml");
