# General Dynamics F-16 Autoflight System
# (c) 2018 Joshua Davidson (it0uchpods)

var max_bank_limit = 30;
setprop("/autopilot/route-manager/advance", 1);

# Every time the waypoint changes, update the stored time
setlistener("/autopilot/route-manager/current-wp", func {
	setprop("/autopilot/internal/wp-change-time", getprop("/sim/time/elapsed-sec"));
});

# Calculates the optimum distance from waypoint to begin turning to next waypoint
var apLoop = maketimer(1, func {
	if (getprop("/sim/model/L70/waypoint-smoothing-enabled"))
	{
		if (getprop("/autopilot/route-manager/route/num") > 0 and getprop("/autopilot/route-manager/active") == 1)
		{
			if ((getprop("/autopilot/route-manager/current-wp") + 1) < getprop("/autopilot/route-manager/route/num"))
			{
				gnds_mps = getprop("/velocities/groundspeed-kt") * 0.5144444444444;
				wp_fly_from = getprop("/autopilot/route-manager/current-wp");
				if (wp_fly_from < 0) {
					wp_fly_from = 0;
				}
				current_course = getprop("/autopilot/route-manager/route/wp[" ~ wp_fly_from ~ "]/leg-bearing-true-deg");
				wp_fly_to = getprop("/autopilot/route-manager/current-wp") + 1;
				if (wp_fly_to < 0) {
					wp_fly_to = 0;
				}
				next_course = getprop("/autopilot/route-manager/route/wp[" ~ wp_fly_to ~ "]/leg-bearing-true-deg");

				delta_angle = math.abs(geo.normdeg180(current_course - next_course));
				max_bank = delta_angle * 1.5;
				if (max_bank > max_bank_limit) {
					max_bank = max_bank_limit;
				}
				radius = (gnds_mps * gnds_mps) / (9.81 * math.tan(max_bank / 57.2957795131));
				time = 0.64 * gnds_mps * delta_angle * 0.7 / (360 * math.tan(max_bank / 57.2957795131));
				delta_angle_rad = (180 - delta_angle) / 114.5915590262;
				R = radius/math.sin(delta_angle_rad);
				dist_coeff = delta_angle * -0.011111 + 2;
				if (dist_coeff < 1) {
					dist_coeff = 1;
				}
				turn_dist = math.cos(delta_angle_rad) * R * dist_coeff / 1852;
				if (getprop("/gear/gear[0]/wow") == 1 and turn_dist < 1) {
					turn_dist = 1;
				}
				setprop("/autopilot/route-manager/advance", turn_dist);
				if (getprop("/sim/time/elapsed-sec")-getprop("/autopilot/internal/wp-change-time") > 60) {
					setprop("/autopilot/internal/wp-change-check-period", time);
				}
				
				if (getprop("/autopilot/route-manager/wp/dist") <= turn_dist) {
					setprop("/autopilot/route-manager/current-wp", getprop("/autopilot/route-manager/current-wp") + 1);
				}
			}
		}
	}
});

# Burn Baby Burn
var start = setlistener("/sim/signals/fdm-initialized", func {
	apLoop.start();
	removelistener(start);
});
